/*
 * Firedoclipse
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package firedoclipse;

import java.io.IOException;
import java.util.*;

import org.eclipse.jdt.internal.ui.text.JavaWordFinder;
import org.eclipse.jface.text.*;
import org.eclipse.jface.text.hyperlink.AbstractHyperlinkDetector;

@SuppressWarnings("restriction")
public abstract class DocHyperlinkDetector extends AbstractHyperlinkDetector {
    
    /**
     * Load language keywords from a properties file
     * 
     * @param language language to load
     * @return keyword-to-URL map
     */
    @SuppressWarnings("unchecked")
    protected static Map<String, String> loadKeywords(String language) {
        Properties keywords = new Properties();
        try {
            keywords.load(DocHyperlinkDetector.class.getResourceAsStream("keywords/" + language + ".properties"));
        } catch (IOException ioe) { }
        return new HashMap<String, String>((Map<String, String>)(Map)keywords);
    }
    
    /**
     * Get the region of the valid Java identifier under the cursor, if any
     * 
     * @param document source code
     * @param region offset
     * @return region of surrounding Java identifier, or <code>null</code> if none
     */
    protected IRegion getJavaWordRegion(IDocument document, IRegion region) {
        IRegion wordRegion = JavaWordFinder.findWord(document, region.getOffset());
        if (wordRegion == null || wordRegion.getLength() == 0) return null;
        return wordRegion;
    }
    
    /**
     * Get the region of the string literal under the cursor, if any
     * 
     * @param document source code
     * @param region offset
     * @return region of surrounding string literal, or <code>null</code> if none
     */
    protected IRegion getStringLiteralRegion(IDocument document, IRegion region) {
        try {
            IRegion line = document.getLineInformationOfOffset(region.getOffset());
            
            int quote = 0;
            boolean quoted = false;
            for (int offset = line.getOffset(); offset <= region.getOffset(); offset++) {
                if (document.getChar(offset) == '"') {
                    quote = offset+1;
                    quoted = !quoted;
                }
            }
            if ( ! quoted) return null;
            
            for (int offset = region.getOffset()+1; offset <= line.getOffset() + line.getLength(); offset++) {
                if (document.getChar(offset) == '"')
                    return new Region(quote, offset - quote);
            }
        } catch (BadLocationException ble) { }
        return null;
    }
    
    /**
     * Get the region of the SGML tag name under the cursor, if any
     * 
     * @param document source code
     * @param region offset
     * @return region of surrounding tag name, or <code>null</code> if none
     */
    protected IRegion getSgmlTagRegion(IDocument document, IRegion region) {
        IRegion word = getJavaWordRegion(document, region);
        if (word == null) return null;
        
        try {
            if ( ! (document.get(word.getOffset()-1, 1).equals("<") || document.get(word.getOffset()-2, 2).equals("</")))
                return null;
        } catch (BadLocationException ble) { return null; }
        
        return word;
    }
}
