/*
 * Firedoclipse
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package jsjr;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Arrays;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.junit.*;

public class ConnectorImplTest {
    
    static final int PORT = 3998;
    static final String ID = "Exportable";
    
    private Connector connector;
    private PrintWriter out;
    
    @Before public void setUp() throws IOException {
        connector = new ConnectorImpl(PORT);
        connector.accept();
        out = new PrintWriter(new Socket("localhost", PORT).getOutputStream());
    }

    @After public void tearDown() throws IOException {
        if (connector != null) connector.close();
        Thread.yield(); // socket needs time...
        if (out != null) out.close();
        Thread.yield(); // ... to die
    }
    
    private void send(String string) {
        out.write(string + "\n");
        out.flush();
    }
    
    @Test(timeout=100) public void testRemote() throws InterruptedException {
        send("remote:interface:jsjr.RemoteTestStub&id:RemoteId");
        assertEquals("init(Connection, RemoteId)", RemoteTestStubProxy.calls.take());
    }
    
    @Test(timeout=100) public void testNoArgSerial() throws InterruptedException {
        send("serial:class:jsjr.SerialTestStub&arguments:0");
        assertEquals("init()", SerialTestStub.calls.take());
    }
    
    @Test(timeout=100) public void testOneArgSerial() throws InterruptedException {
        send("str:One string");
        send("serial:class:jsjr.SerialTestStub&arguments:1");
        assertEquals("init(One string)", SerialTestStub.calls.take());
    }
    
    @Test(timeout=100) public void testMultiArgSerial() throws InterruptedException {
        send("str:One");
        send("str:Two");
        send("str:Red");
        send("str:Blue");
        send("serial:class:jsjr.SerialTestStub&arguments:4");
        assertEquals("init(One, Two, Red, Blue)", SerialTestStub.calls.take());
    }
    
    @Test(timeout=100) public void testNullArgSerial() throws InterruptedException {
        send("null");
        send("str:2nd");
        send("null");
        send("null");
        send("serial:class:jsjr.SerialTestStub&arguments:4");
        assertEquals("init(null, 2nd, null, null)", SerialTestStub.calls.take());
    }
    
    @Test(timeout=100) public void testLongArgSerial() throws InterruptedException {
        send("str:One string");
        send("lstr:that just");
        send("lstr:keeps going");
        send("serial:class:jsjr.SerialTestStub&arguments:1");
        assertEquals("init(One string\nthat just\nkeeps going)", SerialTestStub.calls.take());
    }
    
    @Test(timeout=100) public void testNoArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("call:target:" + ID + "&method:noArg&formals:0&arguments:0");
        assertEquals("noArg()", export.calls.take());
    }
    
    @Test(timeout=500) public void testOneArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("str:Hi there");
        send("class:java.lang.String");
        send("call:target:" + ID + "&method:oneArg&formals:1&arguments:1");
        assertEquals("oneArg(Hi there)", export.calls.take());
    }
    
    @Test(timeout=500) public void testMultiArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("str:Hi");
        send("str:there");
        send("class:java.lang.String");
        send("class:java.lang.String");
        send("call:target:" + ID + "&method:multiArg&formals:2&arguments:2");
        assertEquals("multiArg(Hi, there)", export.calls.take());
    }
    
    @Test(timeout=500) public void testNoArgVarArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("str:Hi");
        send("str:there");
        send("str:!");
        send("class:[Ljava.lang.String;");
        send("call:target:" + ID + "&method:noArgVarArg&formals:1&arguments:3");
        assertEquals("noArgVarArg([Hi, there, !])", export.calls.take());
    }
    
    @Test(timeout=500) public void testOneArgVarArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("str:Hi");
        send("str:there");
        send("str:!");
        send("class:java.lang.String");
        send("class:[Ljava.lang.String;");
        send("call:target:" + ID + "&method:oneArgVarArg&formals:2&arguments:3");
        assertEquals("oneArgVarArg(Hi, [there, !])", export.calls.take());
    }
    
    @Test(timeout=500) public void testMultiArgVarArgCall() throws InterruptedException {
        ExportTestStub export = new ExportTestStub();
        connector.export(ID, export);
        send("str:Hi");
        send("str:there");
        send("str:!");
        send("str:Hi");
        send("str:again");
        send("str:.");
        send("class:java.lang.String");
        send("class:java.lang.String");
        send("class:java.lang.String");
        send("class:[Ljava.lang.String;");
        send("call:target:" + ID + "&method:multiArgVarArg&formals:4&arguments:6");
        assertEquals("multiArgVarArg(Hi, there, !, [Hi, again, .])", export.calls.take());
    }
}

class ExportTestStub {
    
    BlockingQueue<String> calls = new LinkedBlockingQueue<String>();
    
    public void noArg() {
        calls.offer("noArg()");
    }
    
    public void oneArg(String s) {
        calls.offer("oneArg("+s+")");
    }
    
    public void multiArg(String s, String t) {
        calls.offer("multiArg("+s+", "+t+")");
    }
    
    public void noArgVarArg(String... s) {
        calls.offer("noArgVarArg("+Arrays.toString(s)+")");
    }
    
    public void oneArgVarArg(String s, String... t) {
        calls.offer("oneArgVarArg("+s+", "+Arrays.toString(t)+")");
    }
    
    public void multiArgVarArg(String a, String b, String c, String... d) {
        calls.offer("multiArgVarArg("+a+", "+b+", "+c+", "+Arrays.toString(d)+")");
    }
}

class RemoteTestStubProxy {
    
    static BlockingQueue<String> calls = new LinkedBlockingQueue<String>();
    
    public RemoteTestStubProxy(Connection conn, String remoteId) {
        calls.offer("init(Connection, "+remoteId+")");
    }
}

class SerialTestStub {
    
    static BlockingQueue<String> calls = new LinkedBlockingQueue<String>();
    
    public SerialTestStub() {
        calls.offer("init()");
    }
    
    public SerialTestStub(String s) {
        calls.offer("init("+s+")");
    }
    
    public SerialTestStub(String a, String b, String c, String d) {
        calls.offer("init("+a+", "+b+", "+c+", "+d+")");
    }
}