/*
 * Firedoclipse
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

const Cc = Components.classes;
const Ci = Components.interfaces;

const EXTENSION_NAME = "jsjr";
const EXTENSION_HUMAN_NAME = "JSJR";

const THIS_DESCRIPTION = "connector";
const THIS_CID = Components.ID("{e3157d20-d1be-11dc-95ff-0800200c9a66}");
const THIS_CONTRACTID = "@csail.mit.edu/jsjr/connector;1";

Log = function() { return Cc["@csail.mit.edu/jsjr/log;1"].getService().wrappedJSObject; }

const EnvironmentId = "@mozilla.org/process/environment;1";
const SocketTransportServiceId = "@mozilla.org/network/socket-transport-service;1";
const ScriptableInStreamId = "@mozilla.org/scriptableinputstream;1"
const InStreamPumpId = "@mozilla.org/network/input-stream-pump;1";
const ServerSocketId = "@mozilla.org/network/server-socket;1";

/*
 * Connector.
 */

function Connector() {
    this.wrappedJSObject = this;
}

Connector.prototype = {
    init: function() {
        //dump("Connector.init()\n");
        this.connected = false;
        
        this.socket = null;
        
        this.received = [ ];
        this.listeners = [ ];
        this.builders = { };
    },
    
    addListener: function(listener) {
        this.listeners.push(listener);
    },
    
    removeListener: function(listener) {
        this.listeners = this.listeners.filter(function(e, i, a) { return (e != listener) });
    },
    
    addBuilder: function(id, builder) {
        this.builders[id] = builder;
    },
    
    asyncConnect: function(port) {
        //dump("Connector.asyncConnect(" + port + ")\n");
        if (this.connected) return;
        
        port = (new Number(port)).valueOf();
        
        var transport = Cc[SocketTransportServiceId]
                             .getService(Ci.nsISocketTransportService)
                                 .createTransport(null, 0, "127.0.0.1", port, null);
        
        var output = transport.openOutputStream(0, 0, 0);
        output.write("hello\n", 6);
        
        var stream = transport.openInputStream(0, 0, 0);
        var input = Cc[ScriptableInStreamId]
                        .createInstance(Ci.nsIScriptableInputStream);
        input.init(stream);
        
        var pump = Cc[InStreamPumpId]
                       .createInstance(Ci.nsIInputStreamPump);
        pump.init(stream, -1, -1, 0, 0, false);
        pump.asyncRead({
            data : "",
            onStartRequest: function(request, context) {
                context.wrappedJSObject.connect(transport, output);
            },
            onStopRequest: function(request, context, status) {
                context.wrappedJSObject.disconnect();
            },
            onDataAvailable: function(request, context, stream, offset, count) {
                this.data += input.read(count);
                for (var end = this.data.indexOf("\n"); end >= 0; end = this.data.indexOf("\n")) {
                    context.wrappedJSObject.input(this.data.substring(0, end));
                    this.data = this.data.substring(end + 1);
                }
            }
        }, this);
        
        this.asyncListen(port);
    },
    
    asyncListen: function (port) {
        if (this.socket) return;
        
        port = (new Number(port)).valueOf();
        
        this.socket = Cc[ServerSocketId].createInstance(Ci.nsIServerSocket);
        this.socket.init(port+1, true, -1);
        
        var connector = this;
        this.socket.asyncListen({
            onSocketAccepted: function(serv, transport) {
                //dump("Connector...onSocketAccepted(...)\n");
                transport.close(0);
                connector.asyncConnect(port);
            },
            
            onStopListening: function(serv, status) {
                connector.socket = null;
            }
        });
    },
    
    connect: function(transport, output) {
        //dump("Connector.connect(" + transport + ", " + output + ")\n");
        if (transport.isAlive() && ! this.connected) {
            this.output = output;
            this.connected = true;
            for each (listener in this.listeners) {
                listener.connect(this);
            }
        }
    },
    
    disconnect: function() {
        //dump("Connector.disconnect()\n");
        if (this.connected) {
            this.connected = false;
            for each (listener in this.listeners) {
                listener.disconnect();
            }
        }
    },
    
    input: function(message) {
        var m;
        
        m = /^str:(.*)$/.exec(message)
        if (m != null) {
            this.received.push(m[1]);
            return;
        }
        
        m = /^lstr:(.*)$/.exec(message)
        if (m != null) {
            this.received.push(this.received.pop() + "\n" + m[1]);
            return;
        }
        
        m = /^serial:class:(.*)&arguments:(\d+)$/.exec(message)
        if (m != null) {
            var args = [ ];
            for (var ii = 0; ii < m[2]; ii++) {
                args.unshift(this.received.pop());
            }
            this.received.push(this.builders[m[1]](args));
            return;
        }
        
        m = /^call:target:(.*)&method:(.*)&arguments:(\d+)$/.exec(message)
        if (m != null) {
            var args = [ ];
            for (var ii = 0; ii < m[3]; ii++) {
                args.unshift(this.received.pop());
            }
            var target = Cc[m[1]].getService().wrappedJSObject;
            target[m[2]].apply(target, args);
            return;
        }
        
        if (message == "null") {
            this.received.push(null);
            return;
        }
        
        if (message == "hello") {
            return;
        }
        
        Log().error("Invalid message: " + message);
    },
    
    call: function(remoteId, methodName, argTypes, methodArgs) {
        var data = "";
        for each (arg in methodArgs) {
            switch (typeof arg) {
            case "string":
                data += "str:" + arg.replace(/\n/g, "\nlstr:") + "\n";
                break;
            case "object":
                data += arg.marshal();
                break;
            default:
                throw "Cannot marshal method argument " + typeof arg;
            }
        }
        for each (type in argTypes) {
            data += "class:" + type + "\n";
        }
        data += "call:target:" + remoteId
              + "&method:" + methodName
              + "&formals:" + argTypes.length
              + "&arguments:" + methodArgs.length + "\n";
        try {
            this.output.write(data, data.length);
        } catch (e) {
            this.disconnect();
        }
    },
    
    QueryInterface: function(iid) {
        if ( ! iid.equals(Ci.nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/*
 * Singleton factory for QIable component, with initialization.
 */
function SingletonFactory() {
    this.instance = null;
}

SingletonFactory.prototype.createInstance = function(outer, iid) {
    if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (this.instance == null) {
        this.instance = new Connector();
        this.instance.init();
    }
    
    return this.instance.QueryInterface(iid);
}

const THE_FACTORY = new SingletonFactory();

/*
 * Module automagically loaded by Firefox.
 */

var Module = new Object();

Module.registerSelf = function(compMgr, fileSpec, location, type) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.registerFactoryLocation(THIS_CID,
                                    EXTENSION_HUMAN_NAME + " " + THIS_DESCRIPTION,
                                    THIS_CONTRACTID,
                                    fileSpec, location, type);
}

Module.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.unregisterFactoryLocation(THIS_CID, fileSpec);
}

Module.getClassObject = function(compMgr, cid, iid) {
    if ( ! iid.equals(Components.interfaces.nsIFactory))
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (cid.equals(THIS_CID))
        return THE_FACTORY;

    throw Components.results.NS_ERROR_NO_INTERFACE;
}

Module.canUnload = function(compMgr) {
    return true;
}

function NSGetModule(compMgr, fileSpec) {
    return Module;
}
